unit IC_CSM;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	IC Control Strip Module
	-----------------------
	This file implements a Control Strip Module (CSM) for switching between
	IC profiles.  The implementation follows fairly standard CSM lines.
*)

interface

	uses
		Types,
		Menus,
		InternetConfig;
		
	(*	Globals
		-------
		We currently only use globals to hold resources in RAM so that
		we don't have to spin up the disk.  All the rest of the global
		state is maintained with IC itself.
	*)
	
	type
		ModuleGlobalsRecord =
			record
				magic : OSType;				// always kIC_CSMCreator
				setsMenu : MenuHandle;		// 'MENU' ID=mSets
				statusIcon : Handle;		// 'ics#' ID=rStatusIcon
				popupIcon : Handle;			// 'ics#' ID=rPopupIcon
				miscStrings : Handle;		// 'STR#' ID=rMiscStrings
			end;
		ModuleGlobalsPtr = ^ModuleGlobalsRecord;
		ModuleGlobalsHandle = ^ModuleGlobalsPtr;

{$main}
	function Main(message : longint;
					globals : ModuleGlobalsHandle;
					var statusRect : Rect;
					statusPort : GrafPtr) : longint;
		(*	See comment in implementation part. *)
		
implementation

	uses
		PascalA4,

		GestaltEqu,
		ControlStrip,
		Icons,
		Resources,
		
		ICDebug,
		ICCommonSubs,
		ICDialogs;

	(* ***** Misc Constants ***** *)

	const
		kIC_CSMCreator = 'ICsD';		// Our creator code.
		
		kModuleWidth = 20;				// Width, in pixel, of our CSM on the strip itself
		
		noCMErr = -6660;				// Fake error code to make our error handling more orthogonal

	{$ifc icDebug}

		function GoodGlobals(globals : ModuleGlobalsHandle) : Boolean;
			(*	This routine is used by various assertions to check that the globals
				are correct.
			*)
		begin
			GoodGlobals := (globals <> nil) & (globals^ <> nil) & (globals^^.magic = kIC_CSMCreator);
		end; (* GoodGlobals *)

	{$endc}
	
	(* ***** Resources ***** *)
	
	const
		// 'DLOG'
		
		rErrorDialog = 128;
			ditOK = 1;
			ditUserItem = 3;
		
		// 'MENU'
		
		mSets = 128;
			iEditPreferences = 1;
			(* - *)
			iFirstSet = 3;

		// 'STR#'
		
		rMiscStrings = 128;
			strHelpString = 1;
			strGenericAction = 2;				strFirstAction = strGenericAction;
			strEditPreferencesAction = 3;
			strSwitchSetsAction = 4;
			strMiscErr = 5;
			strNoCMErr = 6;
			strNoICErr = 7;
			strInsufficientICErr = 8;
			strNoMemoryErr = 9;
			strCantFindHelperErr = 10;
			strCantHackIt = 11;
		
		// 'ics#'
		
		rStatusIcon = 256;
		rPopupIcon = 257;
		
		// I tried using resource ID 129 for rStatusIcon and got the weirdest errors, namely
		// the icon seemed to bits of the Printer Status CSM's icon!  I ran through it in
		// MacsBug, and our resource file was definitely at the front when I grabbed the icon
		// in InitModule.  I did not have time to trace through the system and find out
		// what was going on, so I just changed resource ID.

	(* ***** Standard Utility Routines ***** *)

	function HaveComponentManager: Boolean;
		(*	Returns true if the Component Manager is available on this machine. *)
		var
			response: longint;
	begin
		HaveComponentManager := (Gestalt(gestaltComponentMgr, response) = noErr);
	end; (* HaveComponentManager *)

	(* ***** Core Implementation ***** *)
	
	function BuildSetsMenu(globals : ModuleGlobalsHandle; inst : ICInstance) : OSStatus;
		(*	This routine queries the ICInstance to find out the list of
			profiles (ie Sets) that are currently available, and adds them
			in to the setsMenu (part of our globals).  It also checks the
			current profile using the canonical Control Strip check mark
			character.
		*)
		var
			err : OSStatus;
			err2 : OSStatus;
			profileCount : longint;
			profileIndex : longint;
			currentProfile : ICProfileID;
			thisProfile : ICProfileID;
			thisName : Str255;
	begin
		ICAssert( GoodGlobals(globals) );
		ICAssert(globals^^.setsMenu <> nil);
		ICAssert( inst <> nil );
		
		// First normalise the menu by deleting any cruft that it might
		// have after the dividing line.
		
		while CountMItems(globals^^.setsMenu) >= iFirstSet do begin
			DeleteMenuItem(globals^^.setsMenu, iFirstSet);
		end; (* while *)
		
		// Then start a Begin/End sequence to IC, which is required
		// for some of these calls, but makes the entire sequence more
		// efficient regardless.
		
		err := ICBegin(inst, icReadOnlyPerm);
		if err = noErr then begin
		
			// Get and remember the current profile.
			
			err := ICGetCurrentProfile(inst, currentProfile);
			
			// Loop through each of the profiles...
			
			if err = noErr then begin
				err := ICCountProfiles(inst, profileCount);
			end; (* if *)
			if err = noErr then begin
				for profileIndex := 1 to profileCount do begin
					err := ICGetIndProfile(inst, profileIndex, thisProfile);
					
					// Get the name of this profile and add it to the menu.
					// Also, if the profile ID matches the current profile ID,
					// put the appropriate mark in the menu.
					
					if err = noErr then begin
						err := ICGetProfileName(inst, thisProfile, thisName);
					end; (* if *)
					if err = noErr then begin
						if thisName = '' then begin
							thisName := ' ';
						end; (* if *)
						SafeAppendMenu(globals^^.setsMenu, thisName);
						if thisProfile = currentProfile then begin
							SetItemMark(globals^^.setsMenu, CountMItems(globals^^.setsMenu), sdevMenuItemMark);
						end; (* if *)
					end; (* if *)
				end; (* for *)
			end; (* if *)
			
			// Shut down the Begin/End pair.
			
			err2 := ICEnd(inst);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)
	
		BuildSetsMenu := err;
	end; (* BuildSetsMenu *)
	
	function SwitchToSetByIndex(globals : ModuleGlobalsHandle; inst : ICInstance; index : integer) : OSStatus;
		(*	Switch the current profile in the IC instance to be the profile
			with the given index.  We do not remember the profile
			IDs associated with each menu item, so instead we must call
			ICGetIndProfile to get the ID of the index'th one, which we
			then pass to ICSetCurrentProfile.
		*)
		var
			err : OSStatus;
			err2 : OSStatus;
			thisProfile : ICProfileID;
	begin
		{$ifc not icDebug}
			{$unused globals}
		{$endc}
		ICAssert( GoodGlobals(globals) );
		ICAssert(index > 0);
		ICAssert(inst <> nil);
		
		// Start a Begin/End sequence to IC, which is required by
		// the ICGetIndProfile routine.
		
		err := ICBegin(inst, icReadWritePerm);
		if err = noErr then begin
		
			// Call ICGetIndProfile to find the profile ID associated with
			// the index'th profile.
			
			err := ICGetIndProfile(inst, index, thisProfile);
			
			// And then set it as the current profile.
			
			if err = noErr then begin
				err := ICSetCurrentProfile(inst, thisProfile);
			end; (* if *)
			
			// Finally, shut down the Begin/End pair.
			
			err2 := ICEnd(inst);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)
		SwitchToSetByIndex := err;
	end; (* SwitchToSetByIndex *)
	
	(* ***** Standard Control Strip Module Messages ***** *)
	
	procedure CloseModule(globals : ModuleGlobalsHandle);
		(*	Handle the standard CSM close message.  All this does is
			dispose of all the memory we allocated in the InitModule routine.
		*)
		var
			junk : OSStatus;
	begin
		if globals <> nil then begin
			if globals^^.setsMenu <> nil then begin
				DisposeMenu(globals^^.setsMenu);
			end; (* if *)
			if globals^^.statusIcon <> nil then begin
				junk := DisposeIconSuite(globals^^.statusIcon, true);
				ICAssert(junk = noErr);
			end; (* if *)
			if globals^^.popupIcon <> nil then begin
				junk := DisposeIconSuite(globals^^.popupIcon, true);
				ICAssert(junk = noErr);
			end; (* if *)
			if globals^^.miscStrings <> nil then begin
				DisposeHandle(globals^^.miscStrings);
				ICAssert(MemError = noErr);
			end; (* if *)
			DisposeHandle(Handle(globals));
			ICAssert(MemError = noErr);
		end; (* if *)
	end; (* CloseModule *)
	
	function InitModule : longint;
		(*	Handle the standard CSM init message.  The basic idea is to
			allocate our globals and then initialise all the fields.
			Note that Control Strip has established our resource file
			at the top of the chain when it sends us this message, so
			we have access to our resources directly.
		
			See the comment in the MouseClick for an explanation of why
			we don't return an error from this routine (which would cause
			the CSM not to show up on the strip) if IC isn't installed.
		*)
		var
			err : OSStatus;
			globals : ModuleGlobalsHandle;
	begin
		// ICDebugStr('InitModule');
		
		InitICDialogs;
		
		// Allocate our globals.
		
		globals := ModuleGlobalsHandle(NewHandleSysClear(sizeof(ModuleGlobalsRecord)));
		err := CheckMemError(globals);
		
		// Fill out the boring fields and then grab the prototype for the Sets
		// menu.  We keep this in a resource because the first item (Edit Preferences)
		// needs to be localised.
		//
		// Note that we lock down our globals for the duration of this routine 
		// because constructs like globals^^.xxx := SomeThingThatMovesOrPurges();
		// are not safe if the handle is unlocked.
		
		if err = noErr then begin
			HLock(Handle(globals));
			globals^^.magic := kIC_CSMCreator;
			globals^^.setsMenu := GetMenu(mSets);
			err := CheckResError(globals^^.setsMenu);
			if err = noErr then begin

				// Detach the setsMenu, otherwise it gets cleaned up automatically
				// by the Resource Manager when Control Strip closes our resource file.

				DetachResource(Handle(globals^^.setsMenu));
				ICAssert(err = noErr);
			end; (* if *)
		end; (* if *)
		
		// Next get the main status icon suite, using a Control Strip callback
		// which makes sure that all the resources are detached.
		
		if err = noErr then begin
			err := SBGetDetachIconSuite(globals^^.statusIcon, rStatusIcon, svAllSmallData);
		end; (* if *)
		
		// And do the same for our popup icon suite.  Note that most CSM samples
		// use a PICT for this, but we use a 'ics#' because it simplifies the code
		// and it's only 64 bytes vs the PICT's 51 bytes!
		
		if err = noErr then begin
			err := SBGetDetachIconSuite(globals^^.popupIcon, rPopupIcon, svAllSmallData);
		end; (* if *)
		
		// Now get the misc strings 'STR#' resource, and detach it as well.
		
		if err = noErr then begin
			globals^^.miscStrings := Get1Resource('STR#', rMiscStrings);
			err := CheckResError(globals^^.miscStrings);
			if err = noErr then begin
				DetachResource(globals^^.miscStrings);
				ICAssert(ResError = noErr);
			end; (* if *)
		end; (* if *)
		
		// Finally, if we have any errors, we just call our CloseModule routine,
		// which does all the appropriate clean up.
		
		if err <> noErr then begin
			CloseModule(globals);
		end; (* if *)
		
		// Return the result, either a positive globals handle, or a negative error
		// code.  Geez, the Control Strip API is not 32 bit clean, which is pretty sucky
		// IMHO.
		
		if err = noErr then begin
			HUnlock(Handle(globals));
			InitModule := longint(globals);
		end else begin
			InitModule := err;
		end; (* if *)
	end; (* InitModule *)
	
	procedure DrawStatus(globals : ModuleGlobalsHandle; statusRect : Rect);
		(*	Handle the standard CSM draw status message.  This is really easy.
			We just draw the main icon, then draw the poup icon, right next to
			each other.
		*)
		var
			junk : OSStatus;
			iconRect : Rect;
	begin
		iconRect.topLeft := statusRect.topLeft;
		iconRect.right := iconRect.left + 16;
		iconRect.bottom := iconRect.top + 16;
		junk := PlotIconSuite(iconRect, kAlignTopLeft, kTransformNone, globals^^.statusIcon);
		
		iconRect.left := iconRect.right;
		iconRect.right := iconRect.left + 16;
		junk := PlotIconSuite(iconRect, kAlignLeft, kTransformNone, globals^^.popupIcon);
		ICAssert(junk = noErr);
	end; (* DrawStatus *)

	type
		ActionType = (kGenericAction, kEditPreferencesAction, kSwitchSetsAction);

	procedure DefaultButtonUserItemA4 (dlg: DialogPtr; item: integer);
		(*	This is a clone of the routine from ICDialogs that
			does the right thing with A4.
		*)
		var
			itemRect: Rect;
			greyPat : Pattern;
			oldA4 : longint;
	begin
		oldA4 := SetUpA4;
		{$unused item}
		SetPort(dlg);
		GetDItemRect(dlg, ditOK, itemRect);
		PenSize(3, 3);
		InsetRect(itemRect, -4, -4);
		if not GetDControlEnable(dlg, ditOK) then begin

			// Avoid using "qd.gray" so that we can be compiled
			// into a code resource.  Sometimes programming the Mac
			// just sucks.
			
			StuffHex(@greyPat, 'AA55AA55AA55AA55');
			PenPat(greyPat);
		end; (* if *)
		FrameRoundRect(itemRect, 16, 16);
		PenNormal;
		oldA4 := SetA4(oldA4);
	end; (* DefaultButtonUserItemA4 *)

	procedure SetupDefaultButtonUserItemA4 (dlg: DialogPtr; defaultItem, roundRectUserItem: integer);
		(*	This is a clone of the routine from ICDialogs that
			installs DefaultButtonUserItemA4 instead of DefaultButtonUserItem.
		*)
		var
			itemRect: Rect;
	begin
		ICAssert(defaultItem = 1);
		GetDItemRect(dlg, defaultItem, itemRect);
		InsetRect(itemRect, -10, -10);
		SetDItemRect(dlg, roundRectUserItem, itemRect);
		SetDItemHandle(dlg, roundRectUserItem, Handle(@DefaultButtonUserItemA4));
	end; (* SetupDefaultButtonUserItemA4 *)

	procedure DisplayError(globals : ModuleGlobalsHandle; errNum : OSStatus; action : ActionType);
		(*	Display an error in response to a user action.  The routine maps errNum to
			the appropriate error text, calculates the correct action string for the
			given action, and then displays the error alert.  The error alert
			contains text like "Could not ^0 because ^1 (^2)."
		*)
		var
			stringIndex : integer;
			errorString : Str255;
			actionString : Str255;
			ourResFile : integer;
			junk : integer;
			offsetToErrorNumberInErrorString : integer;
			dlg : DialogPtr;
	begin
		// Map the error number to an index into our 'STR#' handle.
		
		case errNum of
			badComponentInstance :
				stringIndex := strNoICErr;
			noCMErr :
				stringIndex := strNoCMErr;
			badComponentSelector :
				stringIndex := strInsufficientICErr;
			memFullErr : 
				stringIndex := strNoMemoryErr;
			afpItemNotFound :
				stringIndex := strCantFindHelperErr;
			noPortErr :
				stringIndex := strCantHackIt;
			otherwise
				stringIndex := strMiscErr;
		end; (* case *)

		// Now get the corresponding error and action strings from the list of strings
		// we keep globally.
		
		SBGetDetachedIndString(@errorString, globals^^.miscStrings, stringIndex);
		SBGetDetachedIndString(@actionString, globals^^.miscStrings, strFirstAction + ord(action));
		
		// If the error string contains an ^2, substitute in the error number.
		
		offsetToErrorNumberInErrorString := Pos('^2', errorString);
		if offsetToErrorNumberInErrorString <> 0 then begin
			Delete(errorString, offsetToErrorNumberInErrorString, length('^2'));
			Insert(DecStr(errNum), errorString, offsetToErrorNumberInErrorString);
		end; (* if *)

		// Now open up our own resource file so that we can display our error alert.
		
		ourResFile := SBOpenModuleResourceFile(kIC_CSMCreator);
		if ourResFile <> -1 then begin
			ParamText(actionString, errorString, '', '');
			
			// Previously I used to just call StopAlert here.
			// However, Control Strip 2.0 runs as a separate background
			// only application, so you can't just put up your own dialogs.
			// Instead, I now create the dialog and then call
			// the Control Strip API routine SBModalDialogInContext
			// to run it in the context of the foreground application.

			dlg := GetNewDialog(rErrorDialog, nil, WindowPtr(-1));
			if dlg <> nil then begin

				SetupDefaultButtonUserItemA4(dlg, ditOK, ditUserItem);
				SBModalDialogInContext(nil, junk);
				DisposeDialog(dlg);

			end else begin
				ICDebugStr('Could not get error dialog.');
				SysBeep(10);
			end; (* if *)

			CloseResFile(ourResFile);
			ICAssert(ResError = noErr);
		end else begin
			ICDebugStr('Could not open our own resource file.');
			SysBeep(10);
		end; (* if *)
	end; (* DisplayError *)

	procedure MouseClick(globals : ModuleGlobalsHandle; statusRect : Rect);
		(*	Handle the standard CSM mouse click message.  We open up a connection
			to IC and rebuild the popup menu based on the current set of profiles.
			Then we call SBTrackpopupMenu to allow the user to choose a set.
			We then switch sets based on the users choice.  We also have to
			handle the Edit Preferences menu choice.
			Finally, we handle error conditions.
		*)
		var
			err : OSStatus;
			err2 : OSStatus;
			action : ActionType;
			item : integer;
			inst : ICInstance;
	begin
		ICAssert(globals^^.setsMenu <> nil);
		
		// Check for the presenc of Component Manager before continuing.
		// I don't think Control Strip will run on any machine that doesn't
		// have Component Manager, but I'm just being paranoid.
		
		action := kGenericAction;
		err := noErr;
		if not HaveComponentManager then begin
			err := noCMErr;
		end; (* if *)
		
		if err = noErr then begin
			
			// Open a session to IC.  Note that we do this each time the
			// user clicks, not one at startup time.  This allows a) the user
			// to install the CSM, reboot, then run the IC application,
			// and have it work.  It also means we don't keep an instance open,
			// which makes debugging IC much easier (because I can deregister
			// and reinstall a new build of the component only if there are
			// no instances open).
			
			err := ICStart(inst, kIC_CSMCreator);
			if err = noErr then begin
				err := ICFindConfigFile(inst, 0, nil);
				
				// Build the poup menu, call Control Strip to track the mouse
				// through it, and then respond to the user's choice.
				
				if err = noErr then begin
					err := BuildSetsMenu(globals, inst);
				end; (* if *)
				if err = noErr then begin
					item := SBTrackpopupMenu(statusRect, globals^^.setsMenu);
					case item of
						0 :
							(* do nothing *) ;
						iEditPreferences :
							begin
								action := kEditPreferencesAction;
								err := ICEditPreferences(inst, "");
							end;
						otherwise
							begin
								action := kSwitchSetsAction;
								err := SwitchToSetByIndex(globals, inst, item - iFirstSet + 1);
							end;
					end; (* case *)
				end; (* if *)
			
				// Shut down our session to IC.
				
				err2 := ICStop(inst);
				if err = noErr then begin
					err := err2;
				end; (* if *)
			end; (* if *)
		end; (* if *)
		
		// Handle any errors that arose.
		
		if err <> noErr then begin
			DisplayError(globals, err, action);
		end; (* if *)
	end; (* MouseClick *)
	
	procedure ShowBalloonHelp(globals : ModuleGlobalsHandle; statusRect : Rect);
		(*	Handle the standard CSM show balloon help message.  We get the masseg
			out of our globals and show it using the Control Strip callback.
		*)
		var
			helpString : Str255;
			junk : OSStatus;
	begin
		SBGetDetachedIndString(@helpString, globals^^.miscStrings, strHelpString);
		junk := SBShowHelpString(statusRect, @helpString);
		
		// Do not assert noErr because you can get hmBalloonAborted
		// error back.
		//
		// ICAssert(junk = noErr);
		
	end; (* ShowBalloonHelp *)
	
	(* ***** Control Strip Module Main Line ***** *)

	function Main(message : longint;
					globals : ModuleGlobalsHandle;
					var statusRect : Rect;
					statusPort : GrafPtr) : longint;
		(*	The standard Control Strip main line.  See the documentation:
		
				<http://devworld.apple.com/ngs/lpp/adrpub/docs/dev/technotes/os/os_06.html>
				
			for details of the parameters.  This is basically a big case statement
			based on what message we receive.
		*)
		var
			oldA4 : longint;
			oldZone : THz;
			result : longint;
	begin
		{$unused statusPort}
		
		// Switch in our A4 world.
		
		oldA4 := SetCurrentA4;
		RememberA4;
		
		// Some heap paranoia.  Control Strip should not be calling us with 
		// the zone set incorrectly.  I previously had an ICAssert(GetZone = SystemZone)
		// here, but it fired under a beta version of Control Strip 2.0, so
		// I replaced it with a zone switch sequence.
		//
		// The Control Strip bug [Radar ID 2215990] has been fixed, so I've put
		// the assert back in, just to verify the bug.  But I decided not to take
		// the paranoia code out.
		
		ICAssert(GetZone = SystemZone);
		
		oldZone := GetZone;
		SetZone(SystemZone);
		
		{$ifc icDebug}
			if message <> sdevInitModule then begin
				ICAssert( GoodGlobals(globals) );
			end; (* if *)
		{$endc}
		
		result := 0;
		
		case message of
			sdevInitModule :
				result := InitModule;
			
			sdevCloseModule :
				CloseModule(globals);
				
			sdevFeatures :
				result := bsl(1, sdevWantMouseClicks) + 
							bsl(1, sdevDontAutoTrack) +
							bsl(1, sdevHasCustomHelp);

			sdevGetDisplayWidth :
				result := kModuleWidth;
				
			sdevPeriodicTickle :
				(* do nothing, result is always set to 0 by default *) ;
				
			sdevDrawStatus :
				DrawStatus(globals, statusRect);
				// No need to set result to 0.
				
			sdevMouseClick :
				MouseClick(globals, statusRect);
				// No need to set result to 0.
				
			sdevSaveSettings :
				(* do nothing, result is always set to 0 by default *) ;
				
			sdevShowBalloonHelp :
				ShowBalloonHelp(globals, statusRect);
				// No need to set result to 0.
			
			otherwise
				ICDebugStr('Unrecognised selector.');
		end; (* case *)

		// Restore environment.
		
		SetZone(oldZone);
		oldA4 := SetA4(oldA4);
		
		Main := result;
	end; (* Main *)

end. (* IC_CSM *)